/*******************************************************************************

  Intel Data Center Bridging (DCB) Software
  Copyright(c) 2007-2010 Intel Corporation.

  This program is free software; you can redistribute it and/or modify it
  under the terms and conditions of the GNU General Public License,
  version 2, as published by the Free Software Foundation.

  This program is distributed in the hope it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
  more details.

  You should have received a copy of the GNU General Public License along with
  this program; if not, write to the Free Software Foundation, Inc.,
  51 Franklin St - Fifth Floor, Boston, MA 02110-1301 USA.

  The full GNU General Public License is included in this distribution in
  the file called "COPYING".

  Contact Information:
  e1000-eedc Mailing List <e1000-eedc@lists.sourceforge.net>
  Intel Corporation, 5200 N.E. Elam Young Parkway, Hillsboro, OR 97124-6497

*******************************************************************************/

#include "includes.h"
#include "common.h"
#include <stdio.h>
#include <syslog.h>
#include <sys/un.h>
#include <sys/stat.h>
#include "lldp_mod.h"
#include "lldptool.h"
#include "lldp.h"
#include "lldp_dcbx.h"
#include "lldp_dcbx_clif.h"

void print_dcbx(u16, char *info);
u32 dcbx_lookup_tlv_name(char *tlvid_str);
int dcbx_print_help();

static const struct lldp_mod_ops dcbx_ops_clif = {
	.lldp_mod_register 	= dcbx_cli_register,
	.lldp_mod_unregister 	= dcbx_cli_unregister,
	.print_tlv		= dcbx_print_tlv,
	.lookup_tlv_name	= dcbx_lookup_tlv_name,
	.print_help		= dcbx_print_help,
};

struct type_name_info dcbx_tlv_names[] = {
	{ (OUI_CEE_DCBX << 8) | 1, "PRE-CEE (ver 1) DCBX TLV",	
		"DCBXv1", print_dcbx },
	{ (OUI_CEE_DCBX << 8) | 2, "CEE DCBX TLV",
		"DCBX", print_dcbx },
	{ INVALID_TLVID,	  NULL, NULL }
};

int dcbx_print_help()
{
	int i;
	struct type_name_info *tn = &dcbx_tlv_names[0];

	while (tn->type != INVALID_TLVID) {
		if (tn->key && strlen(tn->key) && tn->name) {
			printf("   %s", tn->key);
			if (strlen(tn->key)+3 <= 8)
				printf("\t");
			printf("\t: %s\n", tn->name);
		}
		tn++;
	}
	
	return 0;
}

struct lldp_module *dcbx_cli_register(void)
{
	struct lldp_module *mod;

	mod = malloc(sizeof(*mod));
	if (!mod) {
		fprintf(stderr, "failed to malloc module data\n");
		goto out_err;
	}
 	mod->id = LLDP_MOD_DCBX;
	mod->ops = &dcbx_ops_clif;

	return mod;
out_err:
	return NULL;
}

void dcbx_cli_unregister(struct lldp_module *mod)
{
	free(mod);
}

void print_dcbx(u16 len, char *info)
{
	printf("%*.*s\n", 2*len, 2*len, info);
}

/* return 1: if it printed the TLV
 *        0: if it did not
 */
int dcbx_print_tlv(u32 tlvid, u16 len, u8 *info)
{
	int i;
	struct type_name_info *tn = &dcbx_tlv_names[0];

	while (tn->type != INVALID_TLVID) {
		if (tlvid == tn->type) {
			printf("%s\n", tn->name);
			if (tn->print_info) {
				printf("\t");
				tn->print_info(len-4, info);
			}
			return 1;
		}
		tn++;
	}
	
	return 0;
}

u32 dcbx_lookup_tlv_name(char *tlvid_str)
{
	struct type_name_info *tn = &dcbx_tlv_names[0];

	while (tn->type != INVALID_TLVID) {
		if (!strcasecmp(tn->key, tlvid_str))
			return tn->type;
		tn++;
	}
	return INVALID_TLVID;
}
