/*******************************************************************************

  Intel Data Center Bridging (DCB) Software
  Copyright(c) 2007-2010 Intel Corporation.

  This program is free software; you can redistribute it and/or modify it
  under the terms and conditions of the GNU General Public License,
  version 2, as published by the Free Software Foundation.

  This program is distributed in the hope it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
  more details.

  You should have received a copy of the GNU General Public License along with
  this program; if not, write to the Free Software Foundation, Inc.,
  51 Franklin St - Fifth Floor, Boston, MA 02110-1301 USA.

  The full GNU General Public License is included in this distribution in
  the file called "COPYING".

  Contact Information:
  e1000-eedc Mailing List <e1000-eedc@lists.sourceforge.net>
  Intel Corporation, 5200 N.E. Elam Young Parkway, Hillsboro, OR 97124-6497

*******************************************************************************/

#include "includes.h"
#include "common.h"
#include <stdio.h>
#include <syslog.h>
#include <sys/un.h>
#include <sys/stat.h>
#include "lldp_mod.h"
#include "lldptool.h"
#include "lldp_mand.h"
#include "lldp_mand_clif.h"

void print_port_id(u16, char *info);
void print_chassis_id(u16, char *info);
void print_ttl(u16, char *info);
u32 mand_lookup_tlv_name(char *tlvid_str);
int mand_print_help();

static const struct lldp_mod_ops mand_ops_clif = {
	.lldp_mod_register 	= mand_cli_register,
	.lldp_mod_unregister 	= mand_cli_unregister,
	.print_tlv		= mand_print_tlv,
	.lookup_tlv_name	= mand_lookup_tlv_name,
	.print_help		= mand_print_help,
};

struct type_name_info mand_tlv_names[] = {
	{ 0, "End of LLDPDU TLV", "" },
	{ 1, "Chassis ID TLV",    "chassisID", print_chassis_id },
	{ 2, "Port ID TLV",       "portID",    print_port_id },
	{ 3, "Time to Live TLV",  "TTL",       print_ttl },
	{ INVALID_TLVID,	  NULL, NULL }
};

int mand_print_help()
{
	int i;
	struct type_name_info *tn = &mand_tlv_names[0];

	while (tn->type != INVALID_TLVID) {
		if (tn->key && strlen(tn->key) && tn->name) {
			printf("   %s", tn->key);
			if (strlen(tn->key)+3 <= 8)
				printf("\t");
			printf("\t: %s\n", tn->name);
		}
		tn++;
	}
	
	return 0;
}

struct lldp_module *mand_cli_register(void)
{
	struct lldp_module *mod;

	mod = malloc(sizeof(*mod));
	if (!mod) {
		fprintf(stderr, "failed to malloc module data\n");
		goto out_err;
	}
 	mod->id = LLDP_MOD_MAND;
	mod->ops = &mand_ops_clif;

	return mod;
out_err:
	return NULL;
}

void mand_cli_unregister(struct lldp_module *mod)
{
	free(mod);
}

void print_chassis_id(u16 len, char *info)
{
	printf("%*.*s\n", 2*len, 2*len, info);
}

void print_port_id(u16 len, char *info)
{
	printf("%*.*s\n", 2*len, 2*len, info);
}

void print_ttl(u16 len, char *info)
{
	u16 ttl;

	hexstr2bin(info, (u8 *)&ttl, sizeof(ttl));
	ttl = ntohs(ttl);
	printf("%d\n", ttl);
}

/* return 1: if it printed the TLV
 *        0: if it did not
 */
int mand_print_tlv(u32 tlvid, u16 len, u8 *info)
{
	int i;
	struct type_name_info *tn = &mand_tlv_names[0];

	while (tn->type != INVALID_TLVID) {
		if (tlvid == tn->type) {
			printf("%s\n", tn->name);
			if (tn->print_info) {
				printf("\t");
				tn->print_info(len, info);
			}
			return 1;
		}
		tn++;
	}
	
	return 0;
}

u32 mand_lookup_tlv_name(char *tlvid_str)
{
	struct type_name_info *tn = &mand_tlv_names[0];

	while (tn->type != INVALID_TLVID) {
		if (!strcasecmp(tn->key, tlvid_str))
			return tn->type;
		tn++;
	}
	return INVALID_TLVID;
}
